/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
    o+
    o+     File:         express.h
    o+
    o+     Programmer:   Shaun Flisakowski
    o+     Date:         Aug 9, 1998
    o+
    o+     A high-level view of expressions.
    o+
    o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */

#ifndef    EXPRESS_H
#define    EXPRESS_H

#include <stdlib.h>
#include <iostream.h>

#include "symbol.h"

#include "tree.h"


class Type;

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
enum ExpressionType
{
    ET_VoidExpr=0,           // No value

    ET_Constant,
    ET_Variable,
    ET_FunctionCall,

    ET_AssignExpr,
    ET_RelExpr,

    ET_UnaryExpr,
    ET_BinaryExpr,
    ET_TrinaryExpr,          // x ? y : z

    ET_CastExpr,             // (type) x
    ET_IndexExpr             // x[y]...[]
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
enum ConstantType
{
    CT_Char,
    CT_Int,
    CT_Float,

    CT_Enum,
    CT_String,
    CT_Array            // For an array initializer
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
enum AssignOp
{
    AO_Equal,          //  =

    AO_PlusEql,        // +=
    AO_MinusEql,       // -=
    AO_MultEql,        // *=
    AO_DivEql,         // /=
    AO_ModEql,         // %=

    AO_ShlEql,         // <<=
    AO_ShrEql,         // >>=

    AO_BitNotEql,      // ~=
    AO_BitAndEql,      // &=
    AO_BitXorEql,      // ^=
    AO_BitOrEql        // |=
};

enum RelOp
{
    RO_Equal,          // ==
    RO_NotEqual,       // !=

    RO_Less,           // < 
    RO_LessEql,        // <=
    RO_Grtr,           // > 
    RO_GrtrEql         // >=
};

enum UnaryOp
{
    UO_Plus=0,         // +
    UO_Minus,          // -
    UO_BitNot,         // ~
    UO_Not,            // !

    UO_PreInc,         // ++x
    UO_PreDec,         // --x
    UO_PostInc,        // x++
    UO_PostDec,        // x--

    UO_AddrOf,         // &
    UO_Deref,          // *

    UO_Sizeof          // sizeof
};

enum BinaryOp
{
    BO_Plus,        // +
    BO_Minus,       // -
    BO_Mult,        // *
    BO_Div,         // /
    BO_Mod,         // %

    BO_Shl,         // <<
    BO_Shr,         // >>
    BO_BitAnd,      // &
    BO_BitXor,      // ^
    BO_BitOr,       // |

    BO_And,         // &&
    BO_Or,          // ||

    BO_Comma,       // x,y

    BO_Member,      // x.y
    BO_PtrMember    // x->y

    //BO_Index        // x[y]

    BO_Assign,      // An AssignExpr
    BO_Rel          // A RelExpr
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class Expression
{
  public:
    Expression(ExpressionType et);

    ExpressionType    etype;

  friend ostream& operator<< ( ostream& out, const Expression& expr );
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class Constant : public Expression
{
  public:
    Constant(long val);
    Constant(char val);
    Constant(float val);
    Constant(double val);
    Constant(ConstantType ct);
 
    ConstantType    ctype;

    union
    {
      long    lng;
      char    ch;
      double  doub;
    };

  friend ostream& operator<< ( ostream& out, const Constant& constant );
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class EnumConstant : public Constant
{
  public:
    EnumConstant( Symbol *nme );

    Symbol    *name;

  friend ostream& operator<< ( ostream& out, const EnumConstant& enumConstant );
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class StringConstant : public Constant
{
  public:
    StringConstant(char *str);

    int Length() const;

    char    *buff;    // dynamic string storage

  friend ostream& operator<< ( ostream& out, const StringConstant& strConst );
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class ArrayConstant : public Constant
{
  public:
    ArrayConstant();

    int           nitems;
    Constant    **items;        // Array of whatever, including another array.

  friend ostream& operator<< ( ostream& out, const ArrayConstant& arrayConst );
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class Variable : public Expression
{
  public:
    Variable();

    Symbol    *name;

  friend ostream& operator<< ( ostream& out, const Variable& var );
};

class FunctionCall : public Expression
{
  public:
    FunctionCall( Expression *function );
    
    int AddArgs( treenode *args );

    int AddArg( Expression *arg );

    int NArgs() const;

  friend ostream& operator<< ( ostream& out, const FunctionCall& funcCall );
};

class UnaryExpr : public Expression
{
  public:
    UnaryExpr( UnaryOp op, Expression *expr );

  Expression *Operand();

  friend ostream& operator<< ( ostream& out, const UnaryExpr& unary );
};

class BinaryExpr : public Expression
{
  public:
    BinaryExpr( BinaryOp op, Expression *lExpr, Expression *rExpr );

  Expression *LeftExpr();
  Expression *RightExpr();

  friend ostream& operator<< ( ostream& out, const BinaryExpr& binary );
};

class TrinaryExpr : public Expression
{
  public:
    TrinaryExpr( Expression *condExpr,
                 Expression *trueExpr, Expression *falseExpr );

  Expression *CondExpr();
  Expression *TrueExpr();
  Expression *FalseExpr();

  friend ostream& operator<< ( ostream& out, const TrinaryExpr& trinary );
};

class AssignExpr : public BinaryExpr
{
  public:
    AssignExpr( AssignOp op, Expression *lExpr, Expression *rExpr );

    Expression *LValue();
    Expression *RValue();

  friend ostream& operator<< ( ostream& out, const AssignExpr& assign );
};

class RelExpr : public BinaryExpr
{
  public:
    RelExpr( RelOp op, Expression *lExpr, Expression *rExpr );

  friend ostream& operator<< ( ostream& out, const RelExpr& rel );
};

class CastExpr : public UnaryExpr
{
  public:
    CastExpr( Type *typeExpr, Expression *operand );

    Type *type();
};

class IndexExpr : public BinaryExpr
{
  // I should extend this to be able to handle
  // multi-dimensional arrays directly.

  public:
    IndexExpr( Expression *array, Expression *sub );
    
    AddSubscript( Expression *sub );

  Expression *Array();
  Expression *Subscript(int i);

  friend ostream& operator<< ( ostream& out, const IndexExpr& index );
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o

Expression *buildExpr(treenode *node);
Expression *buildIndex(treenode *node);

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o

#endif  /* EXPRESS_H */

